
#include<stdio.h>
#include<stdlib.h>
#include <string.h>
 
// structure of the tree node
struct node{
    int data;
    struct node* left;
    struct node* right;
    int ht;
};
 
// global initialization of root node
struct node* root = NULL;
char arr_lb[2] = {'(', '\0'};
char arr_rb[2] = {')', '\0'};

struct node* CreateNewNode(int);
struct node* Insert(struct node*, int);
struct node* DeleteNode(struct node*, int);
struct node* Search(struct node*, int);
struct node* LeftRotate(struct node*);
struct node* RightRotate(struct node*);
int GetBalance(struct node*);
int GetTreeHeight(struct node*);
void TreeToString(struct node*, char*);
void PrintTree(struct node*);
 

int main(){
    char user_choice;
    int user_data;
    struct node* result = NULL;

    while(1){
        scanf("%c", &user_choice);

        switch(user_choice) {
            case 'i':
                scanf("%d", &user_data);
                root = Insert(root, user_data);
                break;
            case 'd':
                scanf("%d", &user_data);
                // printf("%d:input", user_data);
                result = Search(root, user_data);
                if (result == NULL) printf("False\n");
                else{
                    root = DeleteNode(root, user_data);
                    printf("%d\n", user_data);
                }
                break;
            case 's':
                scanf("%d", &user_data);
                // printf("%d:input", user_data);
                result = Search(root, user_data);
                // printf("%d-->", result->data);
                if (result == NULL) printf("False\n");
                else    printf("True\n");
                break;
            case 'b':
                scanf("%d", &user_data);
                result = Search(root, user_data);
                if (result == NULL) printf("False\n");
                else printf("%d\n", GetBalance(result));
                break;
            case 'p':
                printf("(");
                PrintTree(root);
                printf(")\n");
                break;
            case 'e':
                return 1;
        }
    }
} 

struct node* CreateNewNode(int data){
    struct node* new_node = (struct node*) malloc (sizeof(struct node));
    if (new_node == NULL){
        printf("\nMemory can't be allocated\n");
        return NULL;
    }
    new_node->data = data;
    new_node->left = NULL;
    new_node->right = NULL;
    return new_node;
}
 
struct node* LeftRotate(struct node* root){
    struct node* right_child = root->right;
    root->right = right_child->left;
    right_child->left = root;
    root->ht = GetTreeHeight(root);
    right_child->ht = GetTreeHeight(right_child);
    return right_child;
}
 
struct node* RightRotate(struct node* root){
    struct node* left_child = root->left;
    root->left = left_child->right;
    left_child->right = root;
    root->ht = GetTreeHeight(root);
    left_child->ht = GetTreeHeight(left_child);
    return left_child;
}
 

int GetBalance(struct node* root){
    int lh, rh;
    if (root == NULL)   return 0;
    if (root->left == NULL) lh = 0;
    else    lh = 1 + root->left->ht;
    if (root->right == NULL)    rh = 0;
    else    rh = 1 + root->right->ht;
    return lh - rh;
}
 
int GetTreeHeight(struct node* root){
    int lh, rh;
    if (root == NULL)   return 0;
    if (root->left == NULL) lh = 0;
    else    lh = 1 + root->left->ht;
    if (root->right == NULL)    rh = 0;
    else    rh = 1 + root->right->ht;
    if (lh > rh)    return (lh);
    return (rh);
}
 
struct node* Insert(struct node* root, int data){
    if (root == NULL){
        struct node* new_node = CreateNewNode(data);
        if (new_node == NULL)   return NULL;
        root = new_node;
    }
    else if (data > root->data){
        root->right = Insert(root->right, data);
        if (GetBalance(root) == -2){
            if (data > root->right->data)   root = LeftRotate(root);
            else{
                root->right = RightRotate(root->right);
                root = LeftRotate(root);
            }
        }
    }
    else{
        root->left = Insert(root->left, data);
        if (GetBalance(root) == 2){
            if (data < root->left->data)    root = RightRotate(root);
            else{
                root->left = LeftRotate(root->left);
                root = RightRotate(root);
            }
        }
    }
    root->ht = GetTreeHeight(root);
    return root;
}
 

struct node * MinValueNode(struct node* node){
    struct node* current = node;
    while (current->left != NULL) current = current->left;
    return current;
}

int max(int l, int r){
    if (l>=r) return l;
    return r;
}
struct node * DeleteNode(struct node *root, int x){
    struct node * temp = NULL;
    if (root == NULL)   return NULL;
    if (x > root->data)
        root->right = DeleteNode(root->right, x);
    else if (x < root->data)
        root->left = DeleteNode(root->left, x);
    else{
        if ( (root->left == NULL) || (root->right == NULL) ){
            temp = root->left ? root->left : root->right;
            if (temp == NULL){
                temp = root;
                root = NULL;
            }
            else    *root = *temp;
            free(temp);
        }
        else{
            temp = MinValueNode(root->right);
            root->data = temp->data;
            root->right = DeleteNode(root->right, temp->data);
        }
    }

    if (root == NULL)   return root;
    root->ht = 1 + max(GetTreeHeight(root->left), GetTreeHeight(root->right));
    int curr_balance = GetBalance(root);
    if (curr_balance > 1 && GetBalance(root->left) >= 0)
        return RightRotate(root);
    if (curr_balance > 1 && GetBalance(root->left) < 0){
        root->left = LeftRotate(root->left);
        return RightRotate(root);
    }

    if (curr_balance < -1 && GetBalance(root->right) <= 0)
        return LeftRotate(root);
    if (curr_balance < -1 && GetBalance(root->right) > 0){
        root->right = RightRotate(root->right);
        return LeftRotate(root);
    }
    return root;
}
 

struct node* Search(struct node* root, int key){
    if (root == NULL)   return NULL;
    if(root->data == key)   {return root;}
    if(key > root->data)    Search(root->right, key);
    else    Search(root->left, key);
}



void PrintTree(struct node* root)
{
    if (root == NULL)   return;
    printf("%d", root->data);
    printf("(");
    PrintTree(root->left);
    printf(")(");
    PrintTree(root->right);
    printf(")");
}
// void PrintTree(struct node* root){
//     char str[256];
//     TreeToString(root, str);
//     printf("%s\n", str);
// }
