#include <stdio.h>
#include <stdlib.h>

struct node {
	int data; // data
	int color; // 1-red, 0-black
	struct node* parent; // parent
	struct node* right; // right-child
	struct node* left; // left child
};

struct node* root = NULL;

// function to perform BST insertion of a node
struct node* bst(struct node* trav, struct node* temp){
	if (trav == NULL)	return temp;
	if (temp->data < trav->data){
		trav->left = bst(trav->left, temp);
		trav->left->parent = trav;
	}
	else if (temp->data > trav->data){
		trav->right = bst(trav->right, temp);
		trav->right->parent = trav;
	}
	return trav;
}

// Function performing right rotation of the passed node
void rightrotate(struct node* temp){
	struct node* left = temp->left;
	temp->left = left->right;
	if (temp->left)	temp->left->parent = temp;
	left->parent = temp->parent;
	if (!temp->parent)	root = left;
	else if (temp == temp->parent->left)	temp->parent->left = left;
	else	temp->parent->right = left;
	left->right = temp;
	temp->parent = left;
}

// Function performing left rotation of the passed node
void leftrotate(struct node* temp){
	struct node* right = temp->right;
	temp->right = right->left;
	if (temp->right)	temp->right->parent = temp;
	right->parent = temp->parent;
	if (!temp->parent)	root = right;
	else if (temp == temp->parent->left)	temp->parent->left = right;
	else	temp->parent->right = right;
	right->left = temp;
	temp->parent = right;
}

// This function fixes violations caused by BST insertion
void fixup(struct node* root, struct node* pt){
	struct node* parent_pt = NULL;
	struct node* grand_parent_pt = NULL;

	while ((pt != root) && (pt->color != 0) && (pt->parent->color == 1)){
		parent_pt = pt->parent;
		grand_parent_pt = pt->parent->parent;

		/* Case : A: Parent of pt is left child of Grand-parent of pt */
		if (parent_pt == grand_parent_pt->left){
			struct node* uncle_pt = grand_parent_pt->right;
			/* Case : 1: The uncle of pt is also red Only Recoloring required */
			if (uncle_pt != NULL && uncle_pt->color == 1){
				grand_parent_pt->color = 1;
				parent_pt->color = 0;
				uncle_pt->color = 0;
				pt = grand_parent_pt;
			}
			else {
				/* Case : 2: pt is right child of its parent Left-rotation required */
				if (pt == parent_pt->right) {
					leftrotate(parent_pt);
					pt = parent_pt;
					parent_pt = pt->parent;
				}
				/* Case : 3: pt is left child of its parent Right-rotation required */
				rightrotate(grand_parent_pt);
				int t = parent_pt->color;
				parent_pt->color = grand_parent_pt->color;
				grand_parent_pt->color = t;
				pt = parent_pt;
			}
		}
		/* Case : B: Parent of pt is right child of Grand-parent of pt */
		else {
			struct node* uncle_pt = grand_parent_pt->left;
			/* Case : 1: The uncle of pt is also red Only Recoloring required */
			if ((uncle_pt != NULL) && (uncle_pt->color == 1)){
				grand_parent_pt->color = 1;
				parent_pt->color = 0;
				uncle_pt->color = 0;
				pt = grand_parent_pt;
			}
			else {
				/* Case : 2: pt is left child of its parent Right-rotation required */
				if (pt == parent_pt->left) {
					rightrotate(parent_pt);
					pt = parent_pt;
					parent_pt = pt->parent;
				}
				/* Case : 3: pt is right child of its parent: Left-rotation required */
				leftrotate(grand_parent_pt);
				int t = parent_pt->color;
				parent_pt->color = grand_parent_pt->color;
				grand_parent_pt->color = t;
				pt = parent_pt;
			}
		}
	}
	root->color = 0;
}


// struct node* insert(struct node* root, int data){
//     struct node* newnode = (struct node*)malloc(sizeof(struct node));
//     newnode->right = NULL;
//     newnode->left = NULL;
//     newnode->parent = NULL;
//     newnode->data = data;
//     newnode->color = 1;

//     root = bst(root, newnode);
//     fixup(root, newnode);
//     printf("inserted: %d\tnew root-data: %d\n", data, root->data);
//     return root;
// }

void printtree(struct node* root)
{
    if (root == NULL)   {return;}
    printf("%d %c ", root->data, (root->color == 1)?'R':'B');
    printf("(");
    printtree(root->left);
    printf(")(");
    printtree(root->right);
    printf(")");
}


int main(){

    int n = 7;
    int a[7] = {25,18,50,80,12,100,34 };
 
    // for (int i = 0; i < n; i++) {

    char input[9];
    while (1){
        scanf("%s", input);
        if (input[0] == 't')    break;
        int data = atoi(input);
 
        struct node* temp = (struct node*)malloc(sizeof(struct node));
        temp->right = NULL;
        temp->left = NULL;
        temp->parent = NULL;
        temp->data = data;
        temp->color = 1;


        root = bst(root, temp);
        fixup(root, temp);
        printf("(");
        printtree(root);
        printf(")\n");
    }
 

    // char input[9];
    // while (1){
    //     scanf("%s", input);
    //     if (input[0] == 't')    break;
    //     int data = atoi(input);
    //     insert(root, data);
    //     printtree(root);
    // }
    // printf("---------------------------------------------------------------------------\n");
    // printtree(root);

}